library avimobilelib;

import 'dart:async';
import 'dart:convert';

import 'package:flutter/material.dart';
import 'package:google_api_headers/google_api_headers.dart';
import 'package:google_maps_flutter/google_maps_flutter.dart';
import 'package:geolocator/geolocator.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:uuid/uuid.dart';
import 'package:http/http.dart' as http;
import 'package:crypto/crypto.dart';
import 'package:flutter_google_places/flutter_google_places.dart';
import 'package:google_maps_webservice/places.dart';

/**
 * Avi
 */
class Avi
{
  static final COLOR_PRIMARY = Color(0xfffe4600);

  late BuildContext context;

  Avi(BuildContext context)
  {
    this.context = context;
  }

  Future<void> init(String appKey, String secret, bool sandbox) async {
    _AviConfiguration aviConfiguration = new _AviConfiguration();
    await aviConfiguration.setAppKey(appKey);
    await aviConfiguration.setSecret(secret);
    await aviConfiguration.setSandbox(sandbox);
  }

  void showMainWidget()
  {
    Navigator.push(
      context,
      MaterialPageRoute(builder: (context) => _AviWidget()),
    );
  }

  Future<AviUser?> login(AviUser aviUser) async {
    _AviConnector aviConnector = new _AviConnector();
    await aviConnector._connect("login", _AviConnector.METHOD_POST, null, aviUser.toMap());
    if(aviConnector._status == 200)
    {
      _AviConfiguration aviConfiguration = new _AviConfiguration();

      AviUser aviUser = AviUser.fromMap(aviConnector._responseData["user"]);
      await aviConfiguration.setAviUser(aviUser);
      await aviConfiguration.setToken(aviConnector._responseData["token"]);
      aviUser = (await aviConfiguration.getAviUser())!;

      return aviUser;
    }

    return null;
  }

  Future<AviPlace?> getPlace(double latitude, double longitude) async {
    Map<String, dynamic> parameters = Map<String, dynamic>();
    parameters["latitude"] = latitude.toString();
    parameters["longitude"] = longitude.toString();

    _AviConnector aviConnector = new _AviConnector();
    await aviConnector._connect("place", _AviConnector.METHOD_GET, null, parameters);
    if(aviConnector._status == 200)
    {
      AviPlace aviPlace = AviPlace.fromMap(aviConnector._responseData);
      return aviPlace;
    }
    else
    {
      ScaffoldMessenger.of(context).showSnackBar(SnackBar(
        content: Text(aviConnector._message),
      ));
    }

    return null;
  }

  Future<AviVehicle?> setVehicle(AviVehicle aviVehicle) async {
    _AviConnector aviConnector = new _AviConnector();
    await aviConnector._connect("vehicle", _AviConnector.METHOD_POST, null, aviVehicle.toMap());
    if(aviConnector._status == 200)
    {
      aviVehicle.id = aviConnector._responseData["id"];

      return aviVehicle;
    }
    else
    {
      ScaffoldMessenger.of(context).showSnackBar(SnackBar(
        content: Text(aviConnector._message),
      ));
    }

    return null;
  }

  Future<AviQuote?> getQuote(AviTrip aviTrip) async {
    _AviConnector aviConnector = new _AviConnector();
    Map<String, dynamic> parameters = aviTrip.toMap();
    parameters["questions"] = jsonDecode("[{\"answers\":[{\"answer\":\"Si\",\"id\":1,\"questionId\":1}],\"id\":1,\"parentAnswer\":0,\"parentId\":0,\"question\":\"¿Se puede poner en neutral?\",\"service\":1,\"type\":1},{\"answers\":[{\"answer\":\"Si\",\"id\":1,\"questionId\":2}],\"id\":2,\"parentAnswer\":0,\"parentId\":0,\"question\":\"¿Giran las 4 llantas?\",\"service\":1,\"type\":1},{\"answers\":[{\"answer\":\"A pie de calle\",\"id\":1,\"questionId\":3}],\"id\":3,\"parentAnswer\":0,\"parentId\":0,\"question\":\"¿Donde se encuentra?\",\"service\":1,\"type\":1}]");

    await aviConnector._connect("quote", _AviConnector.METHOD_POST, null, parameters);
    if(aviConnector._status == 200)
    {
      AviQuote aviQuote = AviQuote.fromMap(aviConnector._responseData);
      return aviQuote;
    }
    else
    {
      ScaffoldMessenger.of(context).showSnackBar(SnackBar(
        content: Text(aviConnector._message),
      ));
    }

    return null;
  }

  Future<AviTrip?> requestTrip(AviTrip aviTrip) async {
    _AviConnector aviConnector = new _AviConnector();
    Map<String, dynamic> parameters = aviTrip.toMap();
    parameters["questions"] = jsonDecode("[{\"answers\":[{\"answer\":\"Si\",\"id\":1,\"questionId\":1}],\"id\":1,\"parentAnswer\":0,\"parentId\":0,\"question\":\"¿Se puede poner en neutral?\",\"service\":1,\"type\":1},{\"answers\":[{\"answer\":\"Si\",\"id\":1,\"questionId\":2}],\"id\":2,\"parentAnswer\":0,\"parentId\":0,\"question\":\"¿Giran las 4 llantas?\",\"service\":1,\"type\":1},{\"answers\":[{\"answer\":\"A pie de calle\",\"id\":1,\"questionId\":3}],\"id\":3,\"parentAnswer\":0,\"parentId\":0,\"question\":\"¿Donde se encuentra?\",\"service\":1,\"type\":1}]");

    await aviConnector._connect("trip", _AviConnector.METHOD_POST, null, parameters);
    if(aviConnector._status == 200)
    {
      AviTrip aviTrip = AviTrip.fromMap(aviConnector._responseData);
      return aviTrip;
    }
    else
    {
      ScaffoldMessenger.of(context).showSnackBar(SnackBar(
        content: Text(aviConnector._message),
      ));
    }

    return null;
  }
}


/**
 * AviWidget
 */
class _AviWidget extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return MaterialApp(
      title: 'AVI',
      home: Scaffold(
          appBar: AppBar(
            title: const Text('AVI'),
          ),
          body: _AviRequestServiceWidget()
      ),
    );
  }
}



/**
 * AviRequestServiceWidget
 */
class _AviRequestServiceWidget extends StatefulWidget {
  @override
  _AviRequestServiceWidgetState createState() => _AviRequestServiceWidgetState();
}

class _AviRequestServiceWidgetState extends State<_AviRequestServiceWidget> {
  static final String GOOGLE_KEY = "AIzaSyBVelwMWRmOVy3BcZv0srSq0FrPTXgxYNo";

  Completer<GoogleMapController> _controller = Completer();
  Position? _fromPosition;
  Marker? _fromMarker;
  Position? _toPosition;
  Marker? _toMarker;
  Map<MarkerId, Marker> _markers = <MarkerId, Marker>{};
  bool _fromVisible = false;
  bool _toVisible = false;
  bool _confirmVisible = false;
  String _fromText = "¿Dónde está tu vehículo?";
  String _toText = "¿A dónde vas?";

  static final CameraPosition MEXICO_CITY = CameraPosition(
    target: LatLng(19.42847, -99.12766),
    zoom: 14.4746,
  );

  @override
  Widget build(BuildContext context) {
    return Column(
      mainAxisAlignment: MainAxisAlignment.start,
      children: <Widget>[
        Visibility(
            visible: _fromVisible,
            child: InkWell(
                child: Container(
                  width: double.infinity,
                  margin: EdgeInsets.only(
                      top: 10,
                      left: 10,
                      right: 10,
                      bottom: 5
                  ),
                  padding: EdgeInsets.all(5),
                  decoration: BoxDecoration(
                    borderRadius: BorderRadius.circular(10),
                    color: Color(0xffE6E6E6),
                  ),
                  child: Center(
                    child: RichText(
                      text: TextSpan(
                          text: '$_fromText',
                          style: TextStyle(color: Colors.grey,fontSize: 14)
                      ),
                    ),
                  ),
                ),
                onTap: () => _showFromDialog()
            )
        ),
        Visibility(
            visible: _toVisible,
            child: InkWell(
                child: Container(
                    width: double.infinity,
                    margin: EdgeInsets.only(
                        top: 5,
                        left: 10,
                        right: 10,
                        bottom: 10
                    ),
                    padding: EdgeInsets.all(5),
                    decoration:BoxDecoration(
                      borderRadius:BorderRadius.circular(10),
                      color: Color(0xffE6E6E6),
                    ),
                    child: Center(
                      child: RichText(
                        text: TextSpan(
                            text: '$_toText',
                            style: TextStyle(color: Colors.grey, fontSize:14)
                        ),
                      ),
                    )
                ),
                onTap: () => _showToDialog()
            )
        ),
        Expanded(
            child: GoogleMap(
              mapType: MapType.normal,
              initialCameraPosition: MEXICO_CITY,
              onMapCreated: (GoogleMapController googleMapController) => _onMapCreated(googleMapController),
              myLocationEnabled: true,
              myLocationButtonEnabled: true,
              markers: Set<Marker>.of(_markers.values),
            )
        ),
        Visibility(
          visible: _confirmVisible,
          child: Container(
              width: double.infinity,
              margin: EdgeInsets.all(5),
              child: ElevatedButton(style: ButtonStyle(
                foregroundColor: MaterialStateProperty.all<Color>(Colors.white),
                backgroundColor: MaterialStateProperty.all<Color>(Avi.COLOR_PRIMARY),
              ),
                  onPressed: () => _requestTrip(),
                  child: Text("Confirmar")
              )
          ),
        ),
      ],
    );
  }

  void _onMapCreated(GoogleMapController googleMapController) {
    _controller.complete(googleMapController);
    _getUserLocation();

    _AviConfiguration aviConfiguration = _AviConfiguration();
    aviConfiguration.getAviUser().then((aviUser) {
      if(aviUser != null)
      {
        setState(() {
          _fromVisible = true;
          _toVisible = true;
        });
      }
    });
  }

  void _getUserLocation() async {
    bool serviceEnabled;
    LocationPermission permission;

    // Test if location services are enabled.
    serviceEnabled = await Geolocator.isLocationServiceEnabled();
    if (!serviceEnabled) {
      // Location services are not enabled don't continue
      // accessing the position and request users of the
      // App to enable the location services.
      return Future.error('Location services are disabled.');
    }

    permission = await Geolocator.checkPermission();
    if (permission == LocationPermission.denied) {
      permission = await Geolocator.requestPermission();
      if (permission == LocationPermission.denied) {
        // Permissions are denied, next time you could try
        // requesting permissions again (this is also where
        // Android's shouldShowRequestPermissionRationale
        // returned true. According to Android guidelines
        // your App should show an explanatory UI now.
        return Future.error('Location permissions are denied');
      }
    }

    if (permission == LocationPermission.deniedForever) {
      // Permissions are denied forever, handle appropriately.
      return Future.error(
          'Location permissions are permanently denied, we cannot request permissions.');
    }

    // When we reach here, permissions are granted and we can
    // continue accessing the position of the device.
    Geolocator.getCurrentPosition().then((position) => _setFromPosition(position));
  }

  Future<void> _setFromPosition(Position position) async {
    _controller.future.then((googleMapController) {
      _fromPosition = position;
      googleMapController.moveCamera(CameraUpdate.newCameraPosition(CameraPosition(target: LatLng(_fromPosition!.latitude, _fromPosition!.longitude), zoom: 14.4746)));

      MarkerId markerId = MarkerId("_fromMarker");
      _fromMarker = Marker(
        markerId: markerId,
        position: LatLng(
          _fromPosition!.latitude,
          _fromPosition!.longitude,
        ),
        infoWindow: InfoWindow(title: "Origen", snippet: ''),
        onTap: () {

        },
      );

      if(_markers.containsKey(markerId))
      {
        _markers.remove(markerId);
      }

      setState(() {
        // adding a new marker to map
        _markers[markerId] = _fromMarker!;
      });

      Avi avi = Avi(context);
      avi.getPlace(_fromPosition!.latitude, _fromPosition!.longitude).then((aviPlace) {
        if(aviPlace != null)
        {
          setState(() {
            _fromText = aviPlace.formattedAddress!;
          });
        }
      });
    });
  }

  Future<void> _setToPosition(Position position) async {
    if(_fromPosition == null)
    {
      return;
    }

    _controller.future.then((googleMapController) {
      _toPosition = position;

      LatLngBounds latLngBounds = latLngBoundsBuild([
        LatLng(_fromPosition!.latitude, _fromPosition!.longitude),
        LatLng(_toPosition!.latitude, _toPosition!.longitude)]
      );
      googleMapController.animateCamera(CameraUpdate.newLatLngBounds(
          latLngBounds,
          100
      ));

      MarkerId markerId = MarkerId("_toMarker");
      _toMarker = Marker(
        markerId: markerId,
        position: LatLng(
          _toPosition!.latitude,
          _toPosition!.longitude,
        ),
        infoWindow: InfoWindow(title: "Destino", snippet: ''),
        onTap: () {

        },
      );

      if(_markers.containsKey(markerId))
      {
        _markers.remove(markerId);
      }

      setState(() {
        // adding a new marker to map
        _markers[markerId] = _toMarker!;
      });

      Avi avi = Avi(context);
      avi.getPlace(_toPosition!.latitude, _toPosition!.longitude).then((aviPlace) {
        if(aviPlace != null)
        {
          setState(() {
            _toText = aviPlace.formattedAddress!;
            _confirmVisible = true;
          });
        }
      });
    });
  }

  LatLngBounds latLngBoundsBuild(List<LatLng> list) {
    assert(list.isNotEmpty);
    double? x0, x1, y0, y1;
    for (LatLng latLng in list) {
      if (x0 == null) {
        x0 = x1 = latLng.latitude;
        y0 = y1 = latLng.longitude;
      } else {
        if (latLng.latitude > x1!) x1 = latLng.latitude;
        if (latLng.latitude < x0!) x0 = latLng.latitude;
        if (latLng.longitude > y1!) y1 = latLng.longitude;
        if (latLng.longitude < y0!) y0 = latLng.longitude;
      }
    }
    return LatLngBounds(northeast: LatLng(x1!, y1!), southwest: LatLng(x0!, y0!));
  }

  Future<void> _showFromDialog() async {
    Prediction? prediction = await PlacesAutocomplete.show(
      context: context,
      apiKey: GOOGLE_KEY,
      mode: Mode.overlay,
      //language: "es",
      types: ["address"],
      strictbounds: false,
      location: null,
      radius: null,
      components: [new Component(Component.country, "mx")],
      onError: (e) =>
          print("========================== status: " + e.status + " message:" +
              (e.errorMessage != null ? e.errorMessage! : "")),
    );

    if (prediction != null)
    {
      GoogleMapsPlaces googleMapsPlaces = GoogleMapsPlaces(
        apiKey: GOOGLE_KEY,
        apiHeaders: await GoogleApiHeaders().getHeaders(),
      );

      PlacesDetailsResponse placesDetailsResponse = await googleMapsPlaces.getDetailsByPlaceId(prediction.placeId!);
      Position position = Position(
          longitude: placesDetailsResponse.result.geometry!.location.lng,
          latitude: placesDetailsResponse.result.geometry!.location.lat,
          timestamp: DateTime.now(),
          accuracy: 0,
          altitude: 0,
          heading: 0,
          speed: 0,
          speedAccuracy: 0);

      _setFromPosition(position);
    }
  }

  Future<void> _showToDialog() async {
    Prediction? prediction = await PlacesAutocomplete.show(
      context: context,
      apiKey: GOOGLE_KEY,
      mode: Mode.overlay,
      //language: "es",
      types: ["address"],
      strictbounds: false,
      location: null,
      radius: null,
      components: [new Component(Component.country, "mx")],
      onError: (e) =>
          print("========================== status: " + e.status + " message:" +
              (e.errorMessage != null ? e.errorMessage! : "")),
    );

    if (prediction != null)
    {
      GoogleMapsPlaces googleMapsPlaces = GoogleMapsPlaces(
        apiKey: GOOGLE_KEY,
        apiHeaders: await GoogleApiHeaders().getHeaders(),
      );

      PlacesDetailsResponse placesDetailsResponse = await googleMapsPlaces.getDetailsByPlaceId(prediction.placeId!);
      Position position = Position(
          longitude: placesDetailsResponse.result.geometry!.location.lng,
          latitude: placesDetailsResponse.result.geometry!.location.lat,
          timestamp: DateTime.now(),
          accuracy: 0,
          altitude: 0,
          heading: 0,
          speed: 0,
          speedAccuracy: 0);

      _setToPosition(position);
    }
  }

  void _showWaitingDialog() {
    AlertDialog alertDialog = AlertDialog(
      content: new Row(
        children: [
          CircularProgressIndicator(),
          Container(
              margin: EdgeInsets.only(left: 10),
              child: Text("Espera un momento",
                style:  TextStyle(
                    fontWeight: FontWeight.bold,
                    color: Colors.black,
                    fontSize: 14),
              )
          ),
        ],
      ),
    );

    showDialog(
      barrierDismissible: false,
      context:context,
      builder:(BuildContext context){
        return alertDialog;
      },
    );
  }

  void _showErrorDialog(String error) {
    AlertDialog alertDialog = AlertDialog(
      title: Text("Error"),
      titleTextStyle: TextStyle(fontWeight: FontWeight.bold,color: Colors.red,fontSize: 20),
      actionsOverflowButtonSpacing: 20,
      actions: [
        ElevatedButton(
            onPressed: (){
              Navigator.of(context, rootNavigator: true).pop();
            },
            child: Text("Cerrar")),
      ],
      content: new Row(
        children: [
          Expanded(
              child: Text(error)
          ),
        ],
      ),
    );

    showDialog(
      barrierDismissible: false,
      context:context,
      builder:(BuildContext context){
        return alertDialog;
      },
    );
  }

  void _showMessageDialog(String title, String message, VoidCallback? callback) {
    AlertDialog alertDialog = AlertDialog(
      title: Text(title),
      titleTextStyle: TextStyle(fontWeight: FontWeight.bold,color: Colors.black,fontSize: 20),
      actionsOverflowButtonSpacing: 20,
      actions: [
        ElevatedButton(
            onPressed: (){
              Navigator.of(context, rootNavigator: true).pop();

              if(callback != null)
              {
                callback!();
              }
            },
            child: Text("Aceptar")),
      ],
      content: new Row(
        children: [
          Expanded(
              child: Text(message)
          ),
        ],
      ),
    );

    showDialog(
      barrierDismissible: false,
      context:context,
      builder:(BuildContext context){
        return alertDialog;
      },
    );
  }

  void _requestTrip()
  {
    _showWaitingDialog();

    AviVehicle aviVehicle = AviVehicle();
    aviVehicle.vehicleModelId = 5761;
    aviVehicle.year = 2020;
    aviVehicle.plate = "12345";
    aviVehicle.color = "Azul";

    Avi avi = Avi(context);
    avi.setVehicle(aviVehicle).then((aviVehicle) {
      if(aviVehicle != null)
      {
        print(aviVehicle.toMap().toString());

        AviTrip aviTrip = AviTrip();
        aviTrip.startLatitude = _fromPosition!.latitude;
        aviTrip.startLongitude = _fromPosition!.longitude;
        aviTrip.endLatitude = _toPosition!.latitude;
        aviTrip.endLongitude = _toPosition!.longitude;
        aviTrip.service = AviTrip.SERVICE_TOW_TRUCK;
        aviTrip.vehicleId = aviVehicle.id;

        avi.getQuote(aviTrip).then((aviQuote) {
          Navigator.of(context, rootNavigator: true).pop();

          if(aviQuote != null)
          {
            Dialog confirmDialog = Dialog(
                elevation: 0,
                backgroundColor: Color(0xffffffff),
                shape: RoundedRectangleBorder(
                  borderRadius: BorderRadius.circular(15.0),
                ),
                child: Column(
                    mainAxisAlignment: MainAxisAlignment.start,
                    children: <Widget>[
                      Row(
                        children: [
                          Container(
                            margin: EdgeInsets.only(
                                top: 10,
                                left: 10,
                                right: 10
                            ),
                            child: Text(
                              "Costo estimado",
                              style: TextStyle(
                                  fontWeight: FontWeight.bold,
                                  color: Colors.grey,
                                  fontSize: 12),
                            ),
                          ),
                          Expanded(
                              child: Container(
                                  margin: EdgeInsets.only(
                                      top: 10,
                                      left: 10,
                                      right: 10
                                  ),
                                  child: Text(
                                    "\$" + aviQuote!.cost!.toString(),
                                    textAlign: TextAlign.right,
                                    style: TextStyle(
                                        fontWeight: FontWeight.bold,
                                        color: Colors.black,
                                        fontSize: 16),
                                  )
                              )
                          ),
                        ],
                      ),
                      Center(
                        child: Container(
                          width: 200,
                          margin: EdgeInsets.only(
                              left: 10,
                              right: 10
                          ),
                          child: Expanded(
                              child: Text(
                                "El costo es estimado y puede variar en caso de requerir maniobras o casetas",
                                textAlign: TextAlign.center,
                                style: TextStyle(
                                    fontWeight: FontWeight.bold,
                                    color: Colors.red,
                                    fontSize: 10),
                                overflow: TextOverflow.ellipsis,
                                maxLines: 2,
                              )
                          ),
                        ),
                      ),
                      Row(
                        children: [
                          Container(
                            margin: EdgeInsets.all(10),
                            child: Text(
                              "Servicio",
                              style: TextStyle(
                                  fontWeight: FontWeight.bold,
                                  color: Colors.grey,
                                  fontSize: 12),
                            ),
                          ),
                          Expanded(
                              child: Container(
                                  margin: EdgeInsets.all(10),
                                  child: Text(
                                    "Grúa",
                                    textAlign: TextAlign.right,
                                    style: TextStyle(
                                        fontWeight: FontWeight.bold,
                                        color: Colors.black,
                                        fontSize: 14),
                                  )
                              )
                          ),
                        ],
                      ),
                      Container(
                        alignment: Alignment.topLeft,
                        margin: EdgeInsets.only(
                            left: 10,
                            right: 10
                        ),
                        child: Text(
                          "Origen",
                          style: TextStyle(
                              fontWeight: FontWeight.bold,
                              color: Colors.grey,
                              fontSize: 12),
                        ),
                      ),
                      Center(
                        child: Container(
                          width: 200,
                          margin: EdgeInsets.only(
                              left: 10,
                              right: 10
                          ),
                          child: Expanded(
                              child: Text(
                                _fromText,
                                textAlign: TextAlign.center,
                                style: TextStyle(
                                    color: Colors.black,
                                    fontSize: 12),
                                overflow: TextOverflow.ellipsis,
                                maxLines: 4,
                              )
                          ),
                        ),
                      ),
                      Container(
                        alignment: Alignment.topLeft,
                        margin: EdgeInsets.only(
                            top: 10,
                            left: 10,
                            right: 10
                        ),
                        child: Text(
                          "Destino",
                          style: TextStyle(
                              fontWeight: FontWeight.bold,
                              color: Colors.grey,
                              fontSize: 12),
                        ),
                      ),
                      Center(
                        child: Container(
                          width: 200,
                          margin: EdgeInsets.only(
                              left: 10,
                              right: 10
                          ),
                          child: Expanded(
                              child: Text(
                                _toText,
                                textAlign: TextAlign.center,
                                style: TextStyle(
                                    color: Colors.black,
                                    fontSize: 12),
                                overflow: TextOverflow.ellipsis,
                                maxLines: 4,
                              )
                          ),
                        ),
                      ),
                      Container(
                          margin: EdgeInsets.all(20),
                          child: ElevatedButton(style: ButtonStyle(
                            foregroundColor: MaterialStateProperty.all<Color>(Colors.white),
                            backgroundColor: MaterialStateProperty.all<Color>(Avi.COLOR_PRIMARY),
                          ),
                              onPressed: () {
                                Navigator.of(context, rootNavigator: true).pop();
                                _confirmTrip(aviTrip);
                              },
                              child: Text("Confirmar")
                          )
                      )
                    ]
                )
            );


            showDialog(
              barrierDismissible: false,
              context: context,
              builder: (context) {
                return confirmDialog;
              },
            );
          }
        });
      }
      else
      {
        Navigator.of(context, rootNavigator: true).pop();
      }
    });
  }

  void _confirmTrip(AviTrip aviTrip)
  {
    _showWaitingDialog();

    Avi avi = Avi(context);
    avi.requestTrip(aviTrip).then((aviTrip) {
      Navigator.of(context, rootNavigator: true).pop();

      if(aviTrip != null)
      {
        _showMessageDialog("Aviso", "Tu servicio ha sido reservado correctamente", () {
          Navigator.of(context, rootNavigator: true).pop();
          //Navigator.pop(context);
        });
      }
    });
  }
}

/**
 * _AviConnector
 */
class _AviConnector extends http.BaseClient
{
  final String URL = "https://zazu.aviapp.mx";
  //static final String URL = "http://10.255.255.10/public/avi";
  static final String METHOD_POST = "POST";
  static final String METHOD_GET = "GET";
  static final String METHOD_PUT = "PUT";
  static final String METHOD_DELETE = "DELETE";

  late int _status;
  late String _message;
  late Map<String, dynamic> _responseData;
  late http.Client _httpClient;

  Future<void> _connect(String endpoint, String method, int? id, Map<String, dynamic> parameters) async{
    Uri uri = await _getUrl(endpoint, id, method == METHOD_GET ? parameters : null);
    _httpClient = http.Client();

    print("*******************    endpoint: " + endpoint + " parameters: " + parameters.toString());
    try
    {
      http.Response? response;

      if(method == METHOD_GET)
      {
        response = await get(uri);
      }
      else if(method == METHOD_POST)
      {
        response = await post(uri, body: jsonEncode(parameters));
      }
      else if(method == METHOD_PUT)
      {
        response = await put(uri, body: jsonEncode(parameters));
      }
      else if(method == METHOD_DELETE)
      {
        response = await delete(uri);
      }

      if(response != null)
      {
        _parseResponse(response!);
      }
    }
    catch (ex, stacktrace){
      _status = 0;
      _message = "_AviConnector _connect " + endpoint + " - " + ex.toString();

      print(_message + " " + stacktrace.toString());
    }
    finally{
      close();
    }
  }

  void _parseResponse(http.Response response)
  {
    _status = response.statusCode;
    _responseData = jsonDecode(response.body);

    print("status: " + _status.toString() + " responseData: " + response.body);
    if(_status != 200)
    {
      _message = _responseData["message"];
    }
  }

  Future<Uri> _getUrl(endpoint, int? id, Map<String, dynamic>? parameters) async {
    _AviConfiguration aviConfiguration = new _AviConfiguration();
    bool sandbox = await aviConfiguration.getSandbox();

    String url = URL + (sandbox ? "/sandbox/" : "/") + endpoint;
    if(id != null)
    {
      url = url + "/" + id!.toString();
    }
    if(parameters != null)
    {
      url = url + "?" + Uri(queryParameters: parameters).query;
    }
    print(url);
    return Uri.parse(url);
  }

  @override
  Future<http.StreamedResponse> send(http.BaseRequest request) async {
    _AviConfiguration aviConfiguration = new _AviConfiguration();
    String deviceId = await aviConfiguration.getDeviceId();
    String? token = await aviConfiguration.getToken();
    String appKey = await aviConfiguration.getAppKey();
    String secret = await aviConfiguration.getSecret();

    String fingerprint = "flutter";
    int requestId = DateTime.now().toUtc().millisecondsSinceEpoch;
    String contentDigest = base64.encode(sha256.convert(utf8.encode(request.toString())).bytes);
    String signature = base64.encode(Hmac(sha256, utf8.encode(secret)).convert(utf8.encode(contentDigest + requestId.toString() + appKey)).bytes);

    if(token != null)
    {
      request.headers['x-avi-token'] = token!;
    }
    request.headers['x-avi-app-key'] = appKey;
    request.headers['x-avi-device-uuid'] = deviceId;
    request.headers['x-avi-fingerprint'] = fingerprint;
    request.headers['x-avi-request-id'] = requestId.toString();
    request.headers['x-avi-content-digest'] = contentDigest;
    request.headers['x-avi-signature'] = signature;
    request.headers['x-avi-platform'] = "2";
    request.headers['x-avi-version'] = "2021-04-30";
    request.headers['Accept-Encoding'] = "gzip,deflate";
    request.headers['Connection'] = "keep-alive";


    return _httpClient.send(request);
  }
}



/**
 * _AviConfiguration
 */
class _AviConfiguration
{
  Future<String> getDeviceId() async {
    SharedPreferences sharedPreferences = await SharedPreferences.getInstance();
    String? deviceId = sharedPreferences.getString("_AviConfiguration_deviceId");

    if(deviceId == null || deviceId.length == 0)
    {
      deviceId = Uuid().v4();
      await sharedPreferences.setString("_AviConfiguration_deviceId", deviceId);
    }

    return deviceId!;
  }

  Future<void> setAppKey(String appKey) async {
    SharedPreferences sharedPreferences = await SharedPreferences.getInstance();
    await sharedPreferences.setString("_AviConfiguration_appKey", appKey);
  }

  Future<String> getAppKey() async {
    SharedPreferences sharedPreferences = await SharedPreferences.getInstance();
    String? appKey = sharedPreferences.getString("_AviConfiguration_appKey");

    return appKey!;
  }

  Future<void> setSecret(String secret) async {
    SharedPreferences sharedPreferences = await SharedPreferences.getInstance();
    await sharedPreferences.setString("_AviConfiguration_secret", secret);
  }

  Future<String> getSecret() async {
    SharedPreferences sharedPreferences = await SharedPreferences.getInstance();
    String? secret = sharedPreferences.getString("_AviConfiguration_secret");

    return secret!;
  }

  Future<void> setToken(String? token) async {
    SharedPreferences sharedPreferences = await SharedPreferences.getInstance();
    await sharedPreferences.setString("_AviConfiguration_token", token != null ? token : "");
  }

  Future<String?> getToken() async {
    SharedPreferences sharedPreferences = await SharedPreferences.getInstance();
    String? token = sharedPreferences.getString("_AviConfiguration_token");

    return token;
  }

  Future<void> setAviUser(AviUser? aviUser) async {
    SharedPreferences sharedPreferences = await SharedPreferences.getInstance();
    await sharedPreferences.setString("_AviConfiguration_user", aviUser != null ? jsonEncode(aviUser!.toMap()) : "");
  }

  Future<AviUser?> getAviUser() async {
    SharedPreferences sharedPreferences = await SharedPreferences.getInstance();
    String? userData = sharedPreferences.getString("_AviConfiguration_user");
    if(userData != null && userData!.length > 0)
    {
      AviUser? aviUser = AviUser.fromMap(jsonDecode(userData));
      return aviUser;
    }
    return null;
  }

  Future<void> setSandbox(bool sandbox) async {
    SharedPreferences sharedPreferences = await SharedPreferences.getInstance();
    await sharedPreferences.setBool("_AviConfiguration_sandbox", sandbox);
  }

  Future<bool> getSandbox() async {
    SharedPreferences sharedPreferences = await SharedPreferences.getInstance();
    bool? sandbox = sharedPreferences.getBool("_AviConfiguration_sandbox");

    return sandbox == null ? false : sandbox!;
  }
}



/**
 * Estructuras de datos
 */
class AviUser
{
  static final int TYPE_OWNER = 3;

  late int? id;
  late int? type;
  late String? name;
  late String? lastName;
  late String? telephone;
  late String? email;

  AviUser()
  {
    id = 0;
    type = 0;
    name = "";
    lastName = "";
    telephone = "";
    email = "";
  }

  AviUser.fromMap(Map<String, dynamic> map)
      : id = map['id'],
        type = map['type'],
        name = map['name'],
        lastName = map['lastName'],
        telephone = map['telephone'],
        email = map['email'];

  Map<String, dynamic> toMap() => {
    'id' : id,
    'type' : type,
    'name': name,
    'lastName' : lastName,
    'telephone' : telephone,
    'email': email,
  };
}

class AviPlace
{
  late double latitude;
  late double longitude;
  late String? street;
  late String? number;
  late String? neighborhood;
  late String? state;
  late String? city;
  late String? country;
  late String? zipCode;
  late String? formattedAddress;

  AviPlace()
  {
    latitude = 0;
    longitude = 0;
    street = "";
    number = "";
    neighborhood = "";
    state = "";
    city = "";
    country = "";
    zipCode = "";
    formattedAddress = "";
  }

  AviPlace.fromMap(Map<String, dynamic> map)
      : latitude = map['latitude'],
        longitude = map['longitude'],
        street = map['street'],
        number = map['number'],
        neighborhood = map['neighborhood'],
        state = map['state'],
        city = map['city'],
        country = map['country'],
        zipCode = map['zipCode'],
        formattedAddress = map['formattedAddress']
  ;

  Map<String, dynamic> toMap() => {
    'latitude' : latitude,
    'longitude' : longitude,
    'street': street,
    'number' : number,
    'neighborhood' : neighborhood,
    'state': state,
    'city': city,
    'country': country,
    'zipCode': zipCode,
    'formattedAddress': formattedAddress,
  };
}

class AviRoute
{
  late double? startLatitude;
  late double? startLongitude;
  late double? endLatitude;
  late double? endLongitude;
  late double? bearing;
  late double? distance;
  late int? time;
  //late RouteBounds routeBounds;
  //late ArrayList<String> encodedPoints;
  //late ArrayList<LatLng> routePoints;

  AviRoute()
  {
    startLatitude = 0;
    startLongitude = 0;
    endLatitude = 0;
    endLongitude = 0;
    bearing = 0;
    distance = 0;
    time = 0;
  }

  AviRoute.fromMap(Map<String, dynamic> map)
      : startLatitude = map['startLatitude'],
        startLongitude = map['startLongitude'],
        endLatitude = map['endLatitude'],
        endLongitude = map['endLongitude'],
        bearing = map['bearing'],
        distance = map['distance'],
        time = map['time']
  ;

  Map<String, dynamic> toMap() => {
    'startLatitude' : startLatitude,
    'startLongitude' : startLongitude,
    'endLatitude': endLatitude,
    'endLongitude': endLongitude,
    'bearing': bearing,
    'distance': distance,
    'time': time
  };
}

class AviVehicleBrand
{
  late int? id;
  late String? name;
  late String? image;

  AviVehicleBrand()
  {
    id = 0;
    name = "";
    image = "";
  }

  AviVehicleBrand.fromMap(Map<String, dynamic> map)
      : id = map['id'],
        name = map['name'],
        image = map['image']
  ;

  Map<String, dynamic> toMap() => {
    'id' : id,
    'name' : name,
    'image': image
  };
}

class AviVehicleModel
{
  late int? id;
  late String? name;
  late AviVehicleBrand? vehicleBrand;
  late bool? service;

  AviVehicleModel()
  {
    id = 0;
    name = "";
    vehicleBrand = null;
    service = false;
  }

  AviVehicleModel.fromMap(Map<String, dynamic> map)
      : id = map['id'],
        name = map['name'],
        vehicleBrand = map['vehicleBrand'] != null ? AviVehicleBrand.fromMap(map['vehicleBrand']) : null,
        service = map['service']
  ;

  Map<String, dynamic> toMap() => {
    'id' : id,
    'name' : name,
    'vehicleBrand': vehicleBrand != null ? vehicleBrand!.toMap() : null,
    'service': service
  };
}

class AviVehicle
{
  late int? id;
  late int? year;
  late String? color;
  late String? plate;
  late AviVehicleModel? vehicleModel;
  late int? vehicleModelId;
  late bool? dolly;
  late String? vin;
  late int? bulletproof;
  late String? image;
  late bool? federalPlate;

  AviVehicle()
  {
    id = 0;
    year = 0;
    color = "";
    plate = "";
    vehicleModel = null;
    vehicleModelId = 0;
    dolly = false;
    vin = "";
    bulletproof = 0;
    image = "";
    federalPlate = false;
  }

  AviVehicle.fromMap(Map<String, dynamic> map)
      : id = map['id'],
        year = map['year'],
        color = map['color'],
        plate = map['plate'],
        vehicleModel = map['vehicleModel'] != null ? AviVehicleModel.fromMap(map['vehicleModel']) : null,
        vehicleModelId = map['vehicleModelId'],
        dolly = map['dolly'],
        vin = map['vin'],
        bulletproof = map['bulletproof'],
        image = map['image'],
        federalPlate = map['federalPlate']
  ;

  Map<String, dynamic> toMap() => {
    'id' : id,
    'year' : year,
    'color': color,
    'plate' : plate,
    'vehicleModel' : vehicleModel != null ? vehicleModel!.toMap() : null,
    'vehicleModelId': vehicleModelId,
    'dolly': dolly,
    'vin': vin,
    'bulletproof': bulletproof,
    'image': image,
    'federalPlate': federalPlate,
  };
}

class AviAnswer
{
  late int? id;
  late int? questionId;
  late String? answer;

  AviAnswer()
  {
    id = 0;
    questionId = 0;
    answer = "";
  }

  AviAnswer.fromMap(Map<String, dynamic> map)
      : id = map['id'],
        questionId = map['questionId'],
        answer = map['answer']
  ;

  Map<String, dynamic> toMap() => {
    'id' : id,
    'questionId' : questionId,
    'answer': answer,
  };
}

class AviQuestion
{
  static final int TYPE_SINGLE = 1;
  static final int TYPE_MULTIPLE = 2;
  static final int TYPE_TEXT = 3;

  late int? id;
  late int? service;
  late int? type;
  late int? parentId;
  late int? parentAnswer;
  late String? question;
  late List<AviAnswer> answers;

  AviQuestion()
  {
    id = 0;
    service = 0;
    type = 0;
    parentId = 0;
    parentAnswer = 0;
    question = "";
    answers = [];
  }

  AviQuestion.fromMap(Map<String, dynamic> map)
      : id = map['id'],
        service = map['service'],
        type = map['type'],
        parentId = map['parentId'],
        parentAnswer = map['parentAnswer'],
        question = map['question'],
        answers = map['answers']
  ;

  Map<String, dynamic> toMap() => {
    'id' : id,
    'service' : service,
    'type': type,
    'parentId': parentId,
    'parentAnswer': parentAnswer,
    'question': question,
    'answers': answers,
  };
}

class AviQuote
{
  late double? cost;
  late AviRoute? route;

  AviQuote()
  {
    cost = 0;
    route = null;
  }

  AviQuote.fromMap(Map<String, dynamic> map)
      : cost = map['cost'].toDouble()
  //route = map['route'] != null ? _AviRoute.fromMap(map['route']) : null
      ;

  Map<String, dynamic> toMap() => {
    'cost' : cost,
    //'route' : route != null ? route!.toMap() : null,
  };
}

class AviTrip
{
  static final int SERVICE_TOW_TRUCK = 1;
  static final int SERVICE_BATTERY = 2;
  static final int SERVICE_LOCKSMITH = 3;
  static final int SERVICE_TIRE = 4;
  static final int SERVICE_GAS = 5;

  static final int TYPE_LOCAL = 1;
  static final int TYPE_REMOTE = 2;

  static final int FORM_LOCKSMITH_COMMON = 1;
  static final int FORM_LOCKSMITH_LUXURY = 2;

  static final int FORM_GAS_PREMIUM = 1;
  static final int FORM_GAS_REGULAR = 2;
  static final int FORM_GAS_DIESEL = 3;

  static final int FORM_TOW_TRUCK_DOLLIES = 1;
  static final int FORM_TOW_TRUCK_NO_DOLLIES = 0;

  static final int STATUS_REQUESTED = 1;
  static final int STATUS_ASSIGNED = 2;
  static final int STATUS_ARRIVED = 14;
  static final int STATUS_CONTACT = 16;
  static final int STATUS_CANCELED = 3;
  static final int STATUS_RUNNING = 4;
  static final int STATUS_FINISHED = 5;
  static final int STATUS_DELIVERED = 15;
  static final int STATUS_CANCELED_WITH_COST = 6;

  static final int CANCEL_MOTIVE_CUSTOMER_CARD = 1;
  static final int CANCEL_MOTIVE_CUSTOMER = 2;
  static final int CANCEL_MOTIVE_DRIVER = 3;

  late int? id;
  late String? folio;
  late AviUser? customer;
  late AviUser? driver;
  late AviVehicle? customerVehicle;
  late AviVehicle? serviceVehicle;
  late double? startLatitude;
  late double? startLongitude;
  late double? endLatitude;
  late double? endLongitude;
  late String? startAddress;
  late String? endAddress;
  late double? distance;
  late double? cost;
  late int? duration;
  late String? requestTimestamp;
  late String? assignTimestamp;
  late String? startTimestamp;
  late String? endTimestamp;
  late String? payTimestamp;
  late String? payToOwnerTimestamp;
  late String? cancelTimestamp;
  late int? canceledMotive;
  late int? driverRating;
  late String? driverComments;
  late int? customerRating;
  late String? customerComments;
  late String? image;
  late int? service;
  late int? type;
  late int? form;
  late int? quantity;
  late int? status;
  late List<AviQuestion>? questions;
  late int? vehicleId;

  AviTrip()
  {
    id = 0;
    folio = "";
    customer = null;
    driver = null;
    customerVehicle = null;
    serviceVehicle = null;
    startLatitude = 0;
    startLongitude = 0;
    endLatitude = 0;
    endLongitude = 0;
    startAddress = "";
    endAddress = "";
    distance = 0;
    cost = 0;
    duration = 0;
    requestTimestamp = "";
    assignTimestamp = "";
    startTimestamp = "";
    endTimestamp = "";
    payTimestamp = "";
    payToOwnerTimestamp = "";
    cancelTimestamp = "";
    canceledMotive = 0;
    driverRating = 0;
    driverComments = "";
    customerRating = 0;
    customerComments = "";
    image = "";
    service = 0;
    type = 0;
    form = 0;
    quantity = 0;
    status = 0;
    questions = [];
    vehicleId = 0;
  }

  AviTrip.fromMap(Map<String, dynamic> map)
      : id = map['id'],
        folio = map['folio'],
        customer = map['customer'] != null ? AviUser.fromMap(map['customer']) : null,
        driver = map['driver'] != null ? AviUser.fromMap(map['driver']) : null,
        customerVehicle = map['customerVehicle'] != null ? AviVehicle.fromMap(map['customerVehicle']) : null,
        serviceVehicle = map['serviceVehicle'] != null ? AviVehicle.fromMap(map['serviceVehicle']) : null,
        startLatitude = map['startLatitude'],
        startLongitude = map['startLongitude'],
        endLatitude = map['endLatitude'],
        endLongitude = map['endLongitude'],
        startAddress = map['startAddress'],
        endAddress = map['endAddress'],
        distance = map['distance'].toDouble(),
        cost = map['cost'].toDouble(),
        duration = map['duration'],
        requestTimestamp = map['requestTimestamp'],
        assignTimestamp = map['assignTimestamp'],
        startTimestamp = map['startTimestamp'],
        endTimestamp = map['endTimestamp'],
        payTimestamp = map['payTimestamp'],
        payToOwnerTimestamp = map['payToOwnerTimestamp'],
        cancelTimestamp = map['cancelTimestamp'],
        canceledMotive = map['canceledMotive'],
        driverRating = map['driverRating'],
        driverComments = map['driverComments'],
        customerRating = map['customerRating'],
        customerComments = map['customerComments'],
        image = map['image'],
        service = map['service'],
        type = map['type'],
        form = map['form'],
        quantity = map['quantity'],
        status = map['status'],
        questions = map['questions'],
        vehicleId = map['vehicleId']
  ;

  Map<String, dynamic> toMap() => {
    'id' : id,
    'folio' : folio,
    'customer': customer != null ? customer!.toMap() : null,
    'driver' : driver != null ? driver!.toMap() : null,
    'customerVehicle' : customerVehicle != null ? customerVehicle!.toMap() : null,
    'serviceVehicle': serviceVehicle != null ? serviceVehicle!.toMap() : null,
    'startLatitude': startLatitude,
    'startLongitude': startLongitude,
    'endLatitude': endLatitude,
    'endLongitude': endLongitude,
    'startAddress': startAddress,
    'endAddress': endAddress,
    'distance': distance,
    'cost': cost,
    'duration': duration,
    'requestTimestamp': requestTimestamp,
    'assignTimestamp': assignTimestamp,
    'startTimestamp': startTimestamp,
    'endTimestamp': endTimestamp,
    'payTimestamp': payTimestamp,
    'payToOwnerTimestamp': payToOwnerTimestamp,
    'cancelTimestamp': cancelTimestamp,
    'canceledMotive': canceledMotive,
    'driverRating': driverRating,
    'driverComments': driverComments,
    'customerRating': customerRating,
    'customerComments': customerComments,
    'image': image,
    'service': service,
    'type': type,
    'form': form,
    'quantity': quantity,
    'status': status,
    'questions': questions,
    'vehicleId': vehicleId,
  };
}